#from old_results import *
from coopr.opt.results.results_ import *
import coopr.opt.results.problem
from coopr.opt.results.solver import SolverStatus, TerminationCondition
from coopr.opt.results.problem import ProblemSense
from coopr.opt.results.solution import SolutionStatus, Solution, SolutionMap
from coopr.opt.results.container import *
